/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"
#include "Vector.h"
#include "SrcInfo.h"

#include "../config.h"

#if ENABLE_OPENCV_2_0_0

using namespace std;
using namespace FD;

class CvfdhFeatureSurf;
   
DECLARE_NODE(CvfdhFeatureSurf)
/*Node
 *
 * @name CvfdhFeatureSurf
 * @category FDOPENCV_HARK:Feature
 * @description CvfdhFeatureSurf
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @input_name SRCIN
 * @input_type Vector<ObjectRef>
 * @input_description Source location. A type of an element of the vector is SrcInfo.
 *
 * @output_name FEATOUT
 * @output_type Map<int,ObjectRef>
 * @output_description Output features. The ObjectRef should be Vector<float>.
 *
 END*/
   
   
class CvfdhFeatureSurf : public BufferedNode {

  int m_imageInID;
  int m_srcInID;
  int m_featOutID;
      
public:
  CvfdhFeatureSurf(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params), m_srcInID(-1)
  {
    m_imageInID = addInput("IMAGEIN");
    m_featOutID = addOutput("FEATOUT");
    inOrder = true;
  }      

  virtual int translateInput(string inputName) {
    if (inputName == "SRCIN") {
      m_srcInID = addInput(inputName);
      return m_srcInID;
    }
    else if (inputName == "IMAGEIN") {
      for (unsigned int i = 0; i < inputs.size(); i++) {
	if (inputs[i].name == inputName) {
	  return i;
	}
      }
    }
    else {throw new NodeException(this, inputName+ " is not supported.", __FILE__, __LINE__);}
  }
      
  void calculate(int output_id, int count, Buffer &out)
  {

    RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
    (*(outputs[m_featOutID].buffer))[count] = outputMapPtr;

    RCPtr<Map<int, ObjectRef> > image_map(new Map<int, ObjectRef>);

    // Making the Map<id, IplImage*> for processing
    IplImage* img_in_tmp;
    ObjectRef in = getInput(m_imageInID, count);
    if (typeid(*in) == typeid(CvfdImage)) {
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	return;
      }
      img_in_tmp = cvCloneImage(imagePtr->getImage());
      CvfdImage* image= new CvfdImage(img_in_tmp);  
      (*image_map)[0] = image;
    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {
      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	int store_flag = 1;
	if(m_srcInID != -1){
	  RCPtr<Vector<ObjectRef> > sources_ptr = getInput(m_srcInID, count);
	  const Vector<ObjectRef>& sources = *sources_ptr;
	  store_flag = 0;
	  for (int srci = 0; srci < sources.size(); srci++) {
	    RCPtr<SrcInfo> src = sources[srci];
	    if(src->id == it->first) store_flag = 1;
	  }
	}
	if(store_flag == 0) continue;
	img_in_tmp = cvCloneImage(imagePtr.getImage());
	CvfdImage* image= new CvfdImage(img_in_tmp);  
	(*image_map)[it->first] = image;
      }
    }

    // Main routine
    Map<int, ObjectRef>::iterator img_it = image_map->begin();
    for (; img_it != image_map->end(); ++img_it) {
      IplImage* img_in;      
      CvfdImage& imagePtr = object_cast<CvfdImage>(img_it->second);
      img_in = cvCloneImage(imagePtr.getImage());
    
      RCPtr<Vector<float> > map_contents(new Vector<float>);
    
      // ---------------------------------------------------------------
      // main routine : start

      // (1)load Color Image
      cv::Mat colorImage(img_in);
    
      // (2)convert Color Image to Grayscale for Feature Extraction
      cv::Mat grayImage;
      cvtColor(colorImage, grayImage, CV_BGR2GRAY);
    
      // (3)initialize SURF class
      cv::SURF calc_surf = cv::SURF(500,4,2,true);
    
      // (4)extract SURF
      vector<cv::KeyPoint> kp_vec;
      vector<float> desc_vec;		
      calc_surf(grayImage, cv::Mat(), kp_vec, desc_vec);
    
      // (5)draw keypoints
      cout << "Image Keypoints: " << kp_vec.size() << endl;
#if 1
      vector<cv::KeyPoint>::iterator it = kp_vec.begin(), it_end = kp_vec.end();
      for(; it!=it_end; ++it) {
	// circle(colorImage, Point(it->pt.x, it->pt.y), saturate_cast<int>(it->size*0.25), Scalar(255,255,0));
	(*map_contents).push_back((float)it->pt.x);
	(*map_contents).push_back((float)it->pt.y);
	(*map_contents).push_back((float)it->size*0.25);
      }
#else
      for(int i = 0; i < kp_vec.size(); i++) {
	cv::KeyPoint* point = &(kp_vec[i]);
	// Point center;  // Key Point's Center
	// int radius;      // Radius of Key Point
	// center.x = cvRound(point->pt.x);
	// center.y = cvRound(point->pt.y);
	// radius = cvRound(point->size*0.25);
	// circle(colorImage, center, radius, Scalar(255,255,0), 1, 8, 0);
	(*map_contents).push_back((float)point->pt.x);
	(*map_contents).push_back((float)point->pt.y);
	(*map_contents).push_back((float)point->size*0.25);
      }
#endif

      (*outputMapPtr)[img_it->first] = map_contents;    
    
      // End of the transform code
    
      // IplImage* img_out;
      // img_out = cvCloneImage(img_in);

      // main routine : finish
      // ---------------------------------------------------------------
         
      // CvfdImage* image= new CvfdImage(img_out);       
      // (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage (&img_in);
      // cvReleaseImage (&img_out);

    }

    cvReleaseImage(&img_in_tmp);

  }

      
  NO_ORDER_NODE_SPEEDUP(CvfdhFeatureSurf)
};

#endif
