/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "Exception.h"
#include "CvfdImage.h"

using namespace std;
using namespace FD;

class CvfdCanny;
   
DECLARE_NODE(CvfdCanny)
/*Node
 *
 * @name CvfdCanny
 * @category FDOPENCV_ORG:Morphological Operations
 * @description Implements Canny algorithm for edge detection
 * The function cvCanny finds the edges on the input image image and marks them in the output image edges using the Canny algorithm.
 * The smallest of threshold1 and threshold2 is used for edge linking, the largest - to find initial segments of strong edges.
 *
 * @input_name INPUT
 * @input_description Input image
 * @input_type CvfdImage
 * 
 * @output_name OUTPUT
 * @output_description Image of the edges found by the function.
 * @output_type CvfdImage
 *
 * @parameter_name THRESHOLD1
 * @parameter_type float
 * @parameter_value 50
 * @parameter_description The first threshold
 *
 * @parameter_name THRESHOLD2
 * @parameter_type float
 * @parameter_value 200
 * @parameter_description The second threshold   
 *
 * @parameter_name APERTURE_SIZE
 * @parameter_type int
 * @parameter_value 3
 * @parameter_description Aperture parameter for Sobel operator
 END*/
   
   
class CvfdCanny : public BufferedNode {
  //Input ID
  int m_inputID;
  //Output ID
  int m_outputID;
  //Parameters
  float m_threshold1;
  float m_threshold2;
  int m_aperture_size;
      
public:
  CvfdCanny(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    //add inputs
    m_inputID = addInput("INPUT");
    //add outputs
    m_outputID = addOutput("OUTPUT");
    //Initialize parameters         
    m_threshold1 = dereference_cast<float>(parameters.get("THRESHOLD1"));
    m_threshold2 = dereference_cast<float>(parameters.get("THRESHOLD2"));
    m_aperture_size = dereference_cast<int>(parameters.get("APERTURE_SIZE"));
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {
         
    //Read the inputs
    RCPtr<CvfdImage> imagePtr = getInput(m_inputID,count);  
         
    //Handle
    CvfdImage* imageGray = imagePtr->gray(); 
    CvfdImage* image = new CvfdImage(imageGray); 
         
    int status = cvGetErrMode();
    cvSetErrMode( CV_ErrModeSilent ); 
    __BEGIN__;
    OPENCV_CALL(cvCanny(imageGray->getImage(), image->getImage(), m_threshold1, m_threshold2, m_aperture_size));
    __END__;
    cvSetErrMode( status );
    delete imageGray;         
    if( cvGetErrStatus() != CV_StsOk  )
      {
	throw new GeneralException("OPENCV - Error to apply Canny on the image: " +  CCHAR(cvErrorStr( cvGetErrStatus() )),__FILE__,__LINE__);
      }
         
    out[count] = ObjectRef(image);         
  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdCanny)
};
